from rosy.types import Port, ServerHost

DEFAULT_COORDINATOR_HOST: ServerHost = ''
"""Default coordinator host is empty string, which means all available interfaces."""

DEFAULT_COORDINATOR_PORT: Port = 7679
"""Default coordinator port is 7679, which spells "ROSY" on a phone keypad :)"""
